import os
import sys
import time
from pathlib import Path
from typing import Dict, Union

import requests
from apscheduler.events import EVENT_JOB_ERROR
from apscheduler.schedulers.background import BackgroundScheduler

from think_dashboard_agent import types
from think_dashboard_agent.providers import DEFAULT_PROVIDERS, BaseProvider

__scheduler = BackgroundScheduler()
__inited_providers: Dict[str, BaseProvider] = {}
requests_session = requests.Session()


def __run(
        config: types.Config,
        providers: Dict[str, BaseProvider] = None,
        auto_close: bool = True,
        post_to_api: bool = False,
        session: requests.Session = None
):
    _results = []
    if providers is None:
        providers = DEFAULT_PROVIDERS
    else:
        providers = {**DEFAULT_PROVIDERS, **providers}

    for instance in config.instances:
        if instance.key not in __inited_providers:
            provider = providers[instance.type]
            __inited_providers[instance.key] = provider.init(instance)
    for provider in __inited_providers.values():
        instance_result = provider.check(auto_close).dict()
        _results.append({
            'key': getattr(getattr(provider, 'instance'), 'key'),
            'type': getattr(getattr(provider, 'instance'), 'type'),
            'name': getattr(getattr(provider, 'instance'), 'name'),
            'response': instance_result,
        })

    post_body = {
        'server': config.server_name,
        'instances': _results,
    }
    if post_to_api:
        if session is None:
            session = requests_session
            session.headers.update({'Authorization': f"Token {config.api_key}"})
        try:
            r = session.post(config.dashboard_url, data=post_body)
            if r.status_code != 200:
                raise Exception(f'Failed to post to dashboard: {r.status_code}')
        finally:
            if auto_close is True:
                session.close()
    else:
        print(post_body)


def run(conf_file: Union[str, Path], providers: Dict[str, BaseProvider] = None):
    config = types.Config.from_file(conf_file)

    __run(config, providers, auto_close=True)


def run_forever(
        conf_file: Union[str, Path],
        interval: int = 30,
        providers: Dict[str, BaseProvider] = None,
):
    config = types.Config.from_file(conf_file)

    requests_session.headers.update({'Authorization': f"Token {config.api_key}"})

    __scheduler.add_job(
        __run,
        'interval',
        seconds=interval,
        args=[
            config,
            providers,
            False,
            True,
            requests_session,
        ]
    )
    __scheduler.add_listener(__job_exception_handler, EVENT_JOB_ERROR)
    __scheduler.start()
    print('Press Ctrl+{0} to exit'.format('Break' if os.name == 'nt' else 'C'))
    try:
        while __scheduler.running:
            time.sleep(2)
    except (KeyboardInterrupt, SystemExit):
        print('Shutting down...')
    finally:
        requests_session.close()
        if __scheduler.running:
            __scheduler.shutdown(wait=False)
        for provider in __inited_providers.values():
            provider.close()
        sys.exit(0)


def __job_exception_handler(event):  # noqa
    if __scheduler.running:
        __scheduler.shutdown(wait=False)



import psycopg2

from think_dashboard_agent.types import DatabaseInstance, InstanceCheckResult
from .base import BaseProvider


class DatabaseProvider(BaseProvider):
    def __init__(self, instance: DatabaseInstance):
        self.instance = instance
        self.__connection = None

    @classmethod
    def init(cls, instance: DatabaseInstance) -> "DatabaseProvider":
        return cls(instance)

    # Create a connection to the database
    def connect(self):
        if self.__connection is None or self.__connection.closed > 0:
            self.__connection = psycopg2.connect(
                host=self.instance.host,
                port=self.instance.port,
                user=self.instance.username,
                password=self.instance.password,
                database=self.instance.database
            )
        return self.__connection

    # Close the connection to the database
    def close(self):
        if self.__connection is not None:
            self.__connection.close()
            self.__connection = None

    def exc(self, connection):
        cur = connection.cursor()
        cur.execute("""
                    SELECT count(table_name) 
                        FROM information_schema.tables
                        WHERE table_schema = 'public'
                    """)
        return cur.fetchone()[0]

    def check(self, auto_close: bool = True) -> InstanceCheckResult:
        try:
            connection = self.connect()
            result = self.exc(connection)
            return InstanceCheckResult(data=result)
        except psycopg2.OperationalError:
            return InstanceCheckResult(status=500, error='Unable to connect to the database')
        except psycopg2.Error as e:
            return InstanceCheckResult(status=500, error=str(e))
        finally:
            if auto_close:
                self.close()



import redis

from think_dashboard_agent.types import RedisInstance, InstanceCheckResult
from .base import BaseProvider


class RedisProvider(BaseProvider):
    def __init__(self, instance: RedisInstance):
        self.instance = instance
        self.__connection = None

    @classmethod
    def init(cls, instance: RedisInstance) -> "RedisProvider":
        return cls(instance)

    def connect(self):
        if self.__connection is not None:
            try:
                self.__connection.ping()
            except redis.exceptions.ConnectionError:
                self.__connection = None
        if self.__connection is None:
            self.__connection = redis.Redis(
                host=self.instance.host,
                port=self.instance.port,
                password=self.instance.password,
                db=self.instance.db
            )
        return self.__connection

    def exc(self, connection: redis.Redis) -> dict:
        return connection.execute_command('PING')

    def check(self, auto_close: bool = True) -> InstanceCheckResult:
        try:
            connection = self.connect()
            r = self.exc(connection)
            return InstanceCheckResult(data=r)
        except redis.RedisError as e:
            return InstanceCheckResult(status=500, data={'error': str(e)})
        finally:
            if auto_close:
                self.close()

    def close(self):
        if self.__connection is not None:
            self.__connection.close()
            self.__connection = None



import requests

from think_dashboard_agent.types import ElasticSearchInstance, InstanceCheckResult
from .base import BaseProvider


class ElasticSearchProvider(BaseProvider):
    def __init__(self, instance: ElasticSearchInstance):
        self.instance = instance
        self.__connection = None

    @classmethod
    def init(cls, instance: ElasticSearchInstance) -> "ElasticSearchProvider":
        return cls(instance)

    def connect(self):
        if not self.__connection:
            session = requests.Session()
            session.auth = (self.instance.username, self.instance.password)
            session.headers.update({'Content-Type': 'application/json'})
            self.__connection = session
        return self.__connection

    def close(self):
        if self.__connection:
            self.__connection.close()
            self.__connection = None

    def exc(self, session: requests.Session):
        return session.get(f"{self.instance.host}:{self.instance.port}/_cluster/health")

    def check(self, auto_close: bool = True) -> InstanceCheckResult:
        try:
            session = self.connect()
            r = self.exc(session)
            if r.status_code == 200:
                return InstanceCheckResult(status=200, data=r.json())
            else:
                return InstanceCheckResult(status=r.status_code, error=r.text)
        except requests.exceptions.RequestException as e:
            return InstanceCheckResult(status=500, error=str(e))
        finally:
            if auto_close:
                self.close()
